import struct
import binascii
import timeit
from types import FunctionType
import logging

import crcUtil
import sequence
from snappy import SnapConstantsHelper
from snappy.SnapConstants import SNAP2_ADDR_LEN
from RpcCodec import RpcError, SUPPORTED_PYTON_TYPES, RpcFunc, WrongArgCount, decode_args, encode_args


rcast_dec = struct.Struct(
    "!B3s3sBBBHH")  # Flags | ORIG_SRC 24 | IMM_SRC 24 | SEQ 8 | Distance 8 | Max Distance 8 | Ring Spread 16 | Ring Offset 16
addr_dec = struct.Struct("!3s")

RCAST_NO_FLAG = 0b00000000
RCAST_BROADCAST_FLAG = 0b00000001

log = logging.getLogger("snaplib.%s" % (__name__))

# Override these default settings if you want the optional behavior
rpc_crc = False  # Note that this is now the default value but can be overwritten per RPC by changing it in the RpcDescriptor
validate_rpc_crc = False
promiscuous_crc = False  # Probably only want to enable this if you are a sniffer
validate_rpc_crc_if_present = False


class RcastDescriptor(object):
    def __init__(self, crypto=None):
        self.crypto = crypto
        self.orig_src = None
        self._srcAddr = None
        self.seq = None
        self.distance = 0
        self.max_distance = 255
        self.ring_spread = None
        self.ring_offset = 0
        self.dst_addrs = []
        self.flags = RCAST_NO_FLAG
        self._args = ()
        self.func_name = None
        self.isMulticast = True  # Not applicable to this packet type but helps SNAP Connect checks
        self.dstAddr = '1'  # Not applicable to this packet type but helps SNAP Connect checks
        self.rpc_crc = rpc_crc
        self.recv_crc = None
        self.computed_crc = None

    def get_args(self):
        return self._args

    def set_args(self, args):
        try:
            for arg in args:
                if type(arg) not in SUPPORTED_PYTON_TYPES and not isinstance(arg, RpcFunc):
                    raise TypeError("Unsupported type %s" % type(arg))
                elif isinstance(arg, basestring) and len(arg) > 255:
                    raise TypeError("String length is greater than 255")
                elif isinstance(arg, int) and (arg < -32768 or arg > 0xFFFF):
                    raise TypeError("Integer is out of range")
        except TypeError:
            raise RpcError("Do not receive a list of arguments")

        self._args = args

    def set_srcAddr(self, addr):
        if len(addr) != SNAP2_ADDR_LEN:
            raise RpcError("Source Address must be %i bytes" % (SNAP2_ADDR_LEN))

        intAddr = int(binascii.hexlify(addr), 16)
        if intAddr >= 0xFFFFFF or intAddr <= 0x000000:
            raise RpcError("Source Address must be between \xFF\xFF\xFF and \x00\x00\x00")
        self._srcAddr = addr

    def get_srcAddr(self):
        return self._srcAddr

    def __call__(self, callsink):
        assert callsink
        callsink.remoteAddr = self.srcAddr
        callsink.localNetAddr = 1  # Rcasts don't have groups
        callsink.rpcDescriptor = self
        func = getattr(callsink, self.func_name)
        if callable(func):
            # if __debug__:
            # log.debug("Function %s called on device at %s from %s" % (rpcDescriptor.funcName, \
            # binascii.hexlify(rpcDescriptor.dstAddr), \
            # binascii.hexlify(rpcDescriptor.srcAddr))
            if (isinstance(func, FunctionType) and
                        len(self.args) != func.func_code.co_argcount):
                raise WrongArgCount
            func(*self.args)
        elif __debug__:  # pragma: no cover
            log.error("RpcDescriptor did NOT get a callable object when called")

    def __str__(self):  # pragma: no cover
        return "rcast: %s%s" % (self.func_name, self.args)

    def OnSend(self, sucessful):  # pragma: no cover
        if __debug__ and sucessful:
            self.sendTime = timeit.default_timer()
            pass

    def get_encoded_data(self):
        return encode(self, crypto=self.crypto)

    encoded_data = property(get_encoded_data)
    args = property(get_args, set_args)
    srcAddr = property(get_srcAddr, set_srcAddr)


def encode(rcast_descriptor, crypto=None):
    packet = bytearray(128)
    packet[1] = SnapConstantsHelper.PID_RCAST

    if rcast_descriptor.seq is None:
        seq_num = sequence.get_number('rcast' + rcast_descriptor.srcAddr)
    else:
        seq_num = rcast_descriptor.seq

    rcast_dec.pack_into(packet, 2,
                        rcast_descriptor.flags,
                        rcast_descriptor.orig_src, rcast_descriptor.srcAddr, seq_num,
                        rcast_descriptor.distance, rcast_descriptor.max_distance,
                        rcast_descriptor.ring_spread, rcast_descriptor.ring_offset)
    payload_ofs = rcast_dec.size + 2

    if not (rcast_descriptor.flags & RCAST_BROADCAST_FLAG):
        packet[payload_ofs] = chr(len(rcast_descriptor.dst_addrs))
        payload_ofs += 1
        for addr in rcast_descriptor.dst_addrs:
            addr_dec.pack_into(packet, payload_ofs, addr)
            payload_ofs += 3

    packet[payload_ofs] = chr(len(rcast_descriptor.args))
    args_offset = 3 * (len(rcast_descriptor.args) + 1)
    arg_bytes = encode_args(rcast_descriptor.func_name,
                            (RpcFunc(rcast_descriptor.func_name),) + rcast_descriptor.args,
                            args_offset)

    rpc_crc_val = ""
    if rcast_descriptor.rpc_crc:
        crc16 = crcUtil.getSnapCrc(chr(packet[payload_ofs]) + arg_bytes)
        rpc_crc_val = "%c%c" % (crc16 >> 8, crc16 & 0xff)

    packet[0] = payload_ofs + len(arg_bytes) + len(rpc_crc_val)

    return str(packet[:payload_ofs + 1] + arg_bytes + rpc_crc_val)


def decode(packet_bytes):
    rcast_desc = RcastDescriptor()
    base_payload_ofs = 2

    (rcast_desc.flags,
     rcast_desc.orig_src, rcast_desc.srcAddr, rcast_desc.seq,
     rcast_desc.distance, rcast_desc.max_distance,
     rcast_desc.ring_spread, rcast_desc.ring_offset) = rcast_dec.unpack_from(packet_bytes, base_payload_ofs)
    payload_ofs = base_payload_ofs + rcast_dec.size
    if not (rcast_desc.flags & RCAST_BROADCAST_FLAG):
        len_dst_addrs = ord(packet_bytes[payload_ofs])
        payload_ofs += 1
        for i in xrange(0, len_dst_addrs):
            rcast_desc.dst_addrs.append(packet_bytes[payload_ofs:payload_ofs + 3])
            payload_ofs += 3

    num_args = ord(packet_bytes[payload_ofs])
    args = decode_args(packet_bytes, payload_ofs + 1, num_args)
    rcast_desc.func_name = args.pop(0)
    rcast_desc.args = args

    if validate_rpc_crc or validate_rpc_crc_if_present:
        crc_loc = payload_ofs + 1 + 3 * (num_args + 1)
        if type(rcast_desc.func_name) == str:
            crc_loc += len(rcast_desc.func_name) + 1
        for arg in args:
            if type(arg) == str:
                crc_loc += len(arg) + 1

        if crc_loc < len(packet_bytes) - 1:  # CRC takes up TWO bytes, not just one
            rcast_desc.recv_crc = (ord(packet_bytes[-2]) << 8) + ord(packet_bytes[-1])
            rcast_desc.computed_crc = crcUtil.getSnapCrc(packet_bytes[payload_ofs:-2])
            if not promiscuous_crc and rcast_desc.recv_crc != rcast_desc.computed_crc:
                log.debug("Received RCAST-RPC with invalid appended RPC CRC")
                return None
        elif validate_rpc_crc:
            log.debug("Received RPC packet with too short a payload to have a CRC *and* CRC validation is currently active, packet discarded")
            return None

    return rcast_desc
