import struct
import binascii
import timeit
from types import FunctionType
import logging

import sequence
import crcUtil
from snappy import SnapConstantsHelper
from snappy.SnapConstants import SNAP2_ADDR_LEN
from RpcCodec import RpcError, SUPPORTED_PYTON_TYPES, RpcFunc, WrongArgCount, decode_args, encode_args


srpc_dec = struct.Struct("!3sBBB")  # ORIG_SRC 24 | NEXT_HOP_INDEX 8 | SEQ 8 | ADDRS_LEN 8
addr_dec = struct.Struct("!3s")

log = logging.getLogger("snaplib.%s" % (__name__))

# Override these default settings if you want the optional behavior
rpc_crc = False  # Note that this is now the default value but can be overwritten per RPC by changing it in the RpcDescriptor
validate_rpc_crc = False
promiscuous_crc = False  # Probably only want to enable this if you are a sniffer
validate_rpc_crc_if_present = False


class SrpcDescriptor(object):
    def __init__(self, crypto=None):
        self.crypto = crypto
        self.orig_src = None
        self._srcAddr = None
        self.seq = None
        self.addrs = []
        self.next_hop_index = 0
        self._args = ()
        self.func_name = None
        self.rpc_crc = rpc_crc

    def get_args(self):
        return self._args

    def set_args(self, args):
        try:
            for arg in args:
                if type(arg) not in SUPPORTED_PYTON_TYPES and not isinstance(arg, RpcFunc):
                    raise TypeError("Unsupported type %s" % type(arg))
                elif isinstance(arg, basestring) and len(arg) > 255:
                    raise TypeError("String length is greater than 255")
                elif isinstance(arg, int) and (arg < -32768 or arg > 0xFFFF):
                    raise TypeError("Integer is out of range")
        except TypeError:
            raise RpcError("Do not receive a list of arguments")

        self._args = args

    def set_srcAddr(self, addr):
        if len(addr) != SNAP2_ADDR_LEN:
            raise RpcError("Source Address must be %i bytes" % (SNAP2_ADDR_LEN))

        intAddr = int(binascii.hexlify(addr), 16)
        if intAddr >= 0xFFFFFF or intAddr <= 0x000000:
            raise RpcError("Source Address must be between \xFF\xFF\xFF and \x00\x00\x00")
        self._srcAddr = addr

    def get_srcAddr(self):
        return self._srcAddr

    def __call__(self, callsink):
        assert callsink
        # callsink.remoteAddr = self.srcAddr
        # callsink.localNetAddr = 1  # Rcasts don't have groups
        callsink.rpcDescriptor = self
        func = getattr(callsink, self.func_name)
        if callable(func):
            # if __debug__:
            # log.debug("Function %s called on device at %s from %s" % (rpcDescriptor.funcName, \
            # binascii.hexlify(rpcDescriptor.dstAddr), \
            # binascii.hexlify(rpcDescriptor.srcAddr))
            if (isinstance(func, FunctionType) and
                        len(self.args) != func.func_code.co_argcount):
                raise WrongArgCount
            func(*self.args)
        elif __debug__:  # pragma: no cover
            log.error("RpcDescriptor did NOT get a callable object when called")

    def __str__(self):  # pragma: no cover
        return "srpc: %s%s" % (self.func_name, self.args)

    def OnSend(self, sucessful):  # pragma: no cover
        if __debug__ and sucessful:
            self.sendTime = timeit.default_timer()
            pass

    def get_encoded_data(self):
        return encode(self, crypto=self.crypto)

    encoded_data = property(get_encoded_data)
    args = property(get_args, set_args)
    srcAddr = property(get_srcAddr, set_srcAddr)


def encode(srpc_descriptor, crypto=None):
    packet = bytearray(128)
    packet[1] = SnapConstantsHelper.PID_SRPC

    if srpc_descriptor.seq is None:
        seq_num = sequence.get_number(srpc_descriptor.srcAddr)
    else:
        seq_num = srpc_descriptor.seq

    srpc_dec.pack_into(packet, 2,
                       srpc_descriptor.orig_src, srpc_descriptor.next_hop_index, seq_num,
                       len(srpc_descriptor.addrs))
    payload_ofs = srpc_dec.size + 2

    for addr in srpc_descriptor.addrs:
        addr_dec.pack_into(packet, payload_ofs, addr)
        payload_ofs += 3

    packet[payload_ofs] = chr(len(srpc_descriptor.args))
    args_offset = 3 * (len(srpc_descriptor.args) + 1)
    arg_bytes = encode_args(srpc_descriptor.func_name,
                            (RpcFunc(srpc_descriptor.func_name),) + srpc_descriptor.args,
                            args_offset)

    rpc_crc_val = ""
    if srpc_descriptor.rpc_crc:
        crc16 = crcUtil.getSnapCrc(chr(packet[payload_ofs]) + arg_bytes)
        rpc_crc_val = "%c%c" % (crc16 >> 8, crc16 & 0xff)

    packet[0] = payload_ofs + len(arg_bytes) + len(rpc_crc_val)

    return str(packet[:payload_ofs + 1] + arg_bytes + rpc_crc_val)


def decode(packet_bytes):
    srpc_descriptor = SrpcDescriptor()
    base_payload_ofs = 2

    (srpc_descriptor.orig_src, srpc_descriptor.next_hop_index,
     srpc_descriptor.seq, addrs_len) = srpc_dec.unpack_from(packet_bytes, base_payload_ofs)
    payload_ofs = base_payload_ofs + srpc_dec.size

    for i in xrange(0, addrs_len):
        srpc_descriptor.addrs.append(packet_bytes[payload_ofs:payload_ofs + 3])
        payload_ofs += 3

    num_args = ord(packet_bytes[payload_ofs])
    args = decode_args(packet_bytes, payload_ofs + 1, ord(packet_bytes[payload_ofs]))
    srpc_descriptor.func_name = args.pop(0)
    srpc_descriptor.args = args

    if validate_rpc_crc or validate_rpc_crc_if_present:
        crc_loc = payload_ofs + 1 + 3 * (num_args + 1)
        for arg in args:
            if type(arg) == str:
                crc_loc += len(arg) + 1
        if type(srpc_descriptor.func_name) == str:
            crc_loc += len(srpc_descriptor.func_name) + 1

        if crc_loc < len(packet_bytes) - 1:  # CRC takes up TWO bytes, not just one
            srpc_descriptor.recv_crc = (ord(packet_bytes[-2]) << 8) + ord(packet_bytes[-1])
            srpc_descriptor.computed_crc = crcUtil.getSnapCrc(packet_bytes[payload_ofs:-2])
            if not promiscuous_crc and srpc_descriptor.recv_crc != srpc_descriptor.computed_crc:
                log.debug("Received RCAST-RPC with invalid appended RPC CRC")
                return None
        elif validate_rpc_crc:
            log.debug("Received RPC packet with too short a payload to have a CRC *and* CRC validation is currently active, packet discarded")
            return None

    return srpc_descriptor
